from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from functools import cached_property
from typing import Optional

from auxiliary import ilen, retain_iter

from math2.misc import series_sum


class Deprec(ABC):
    """Deprec is the abstract base class for all depreciations."""

    def __init__(self, basis: float, salvage: float, life: int):
        self.basis = basis
        self.salvage = salvage
        self.life = life

    def cap_gain(self, market: float) -> float:
        """Obtains the capital gain with respect to the market value.

        :param market: The market value.
        :return: The capital gain.
        """
        return max(0.0, market - self.basis)

    def recap_deprec(self, market: float) -> float:
        """Obtains the recaptured depreciation with respect to the market value.

        :param market: The market value.
        :return: The recaptured depreciation.
        """
        return max(0.0, min(self.basis, market) - self.salvage)

    def loss_on_disp(self, market: float) -> float:
        """Obtains the loss on disposal with respect to the market value.

        :param market: The market value.
        :return: The loss on disposal.
        """
        return max(0.0, self.salvage - market)

    @cached_property
    def books(self) -> Iterator[float]:
        """Calculates the book values throughout its life.

        :return: The book values.
        """
        return (self.book(t) for t in range(self.life + 1))

    @abstractmethod
    def book(self, t: int) -> float:
        """The book value at time t.

        :param t: The time.
        :return: The book value.
        """
        pass

    @abstractmethod
    def amount(self, t: int) -> float:
        """The positive loss of the depreciation at time t.

        :param t: The time.
        :return: The loss.
        """
        pass


class StrLineDeprec(Deprec):
    """StrLineDeprec is the class for straight line depreciations."""

    def book(self, t: int) -> float:
        return self.basis - self.amount() * t

    def amount(self, t: Optional[int] = None) -> float:
        return (self.basis - self.salvage) / self.life


class DeclBalDeprec(Deprec):
    """DeclBalDeprec is the class for declining balance depreciations."""

    @cached_property
    def rate(self) -> float:
        return 1 - (self.salvage / self.basis) ** (1 / self.life)

    def book(self, t: int) -> float:
        return self.basis * (1 - self.rate) ** t

    def amount(self, t: int) -> float:
        return self.book(t - 1) * self.rate

    @classmethod
    def from_rate(cls, basis: float, life: int, rate: float) -> DeclBalDeprec:
        """Constructs the declining balance depreciation from salvage value.

        :param basis: The basis.
        :param life: The life.
        :param rate: The depreciation rate.
        :return: The declining balance depreciation.
        """
        return DeclBalDeprec(basis, basis * (1 - rate) ** life, life)


class DblDeclBalDeprec(Deprec):
    """DblDeclBalDeprec is the class for double declining balance depreciations."""

    def __init__(self, basis: float, salvage: float, life: int, floor: bool = False):
        super().__init__(basis, salvage, life)

        self.floor = floor

    @cached_property
    def rate(self) -> float:
        return 2 / self.life

    def book(self, t: int) -> float:
        book = self.basis * (1 - self.rate) ** t

        return max(self.salvage, book) if self.floor else book

    def amount(self, t: int) -> float:
        return self.book(t - 1) * self.rate


class SYDDeprec(Deprec):
    """SYDDeprec is the class for sum-of-years'-digits depreciations."""

    @cached_property
    def syd(self) -> int:
        """
        :return: The sum of the years' digits.
        """
        return series_sum(self.life)

    def book(self, t: int) -> float:
        return self.basis - (self.syd - series_sum(self.life - t)) / self.syd * (self.basis - self.salvage)

    def amount(self, t: int) -> float:
        return (self.life - t + 1) / self.syd * (self.basis - self.salvage)


class UPDeprec(Deprec):
    """UPDeprec is the class for units of production depreciations."""

    @retain_iter
    def __init__(self, basis: float, salvage: float, prods: Iterable[float]):
        super().__init__(basis, salvage, ilen(prods))

        self.prods = list(prods)

    @cached_property
    def lifetime_prod(self) -> float:
        """
        :return: The lifetime production.
        """
        return sum(self.prods)

    def book(self, t: int) -> float:
        return self.basis - sum(self.prods[:t]) / self.lifetime_prod * (self.basis - self.salvage)

    def amount(self, t: int) -> float:
        return self.prods[t - 1] / self.lifetime_prod * (self.basis - self.salvage)



from collections.abc import Iterable, Sequence
from random import shuffle
from statistics import fmean

from auxiliary import retain_iter, trimmed

from math2.typing import _T


def trimmed_mean(values: Iterable[float], percentage: float) -> float:
    """Calculates the trimmed mean of the values.

    :param values: The values.
    :param percentage: The percentage value.
    :return: The trimmed mean.
    """
    return fmean(trimmed(sorted(values), percentage))


@retain_iter
def range_(values: Iterable[float]) -> float:
    """Calculates the range of the values.

    :param values: The values.
    :return: The range.
    """
    return max(values) - min(values)


def shuffled(values: Iterable[_T]) -> Sequence[_T]:
    """Shuffles the copied values and returns the list.

    :param values: The values.
    :return: The shuffled sequence.
    """
    values = list(values)
    shuffle(values)
    return values



from unittest import main

from auxiliary import ExtTestCase

from math2.misc import bind, frange, interp, linspace, prod, series_sum, sum_


class MiscTestCase(ExtTestCase):
    def test_bind(self) -> None:
        self.assertEqual(bind(1, 0, 2), 1)
        self.assertEqual(bind(-100, 0, 2), 0)
        self.assertEqual(bind(100, 0, 2), 2)
        self.assertRaises(ValueError, bind, 100, 2, 0)

    def test_sum(self) -> None:
        self.assertEqual(sum_(range(6)), 15)
        self.assertEqual(sum_(range(6), 0), 15)
        self.assertEqual(sum_((), 1), 1)
        self.assertRaises(ValueError, sum_, ())

    def test_prod(self) -> None:
        self.assertEqual(prod(range(6)), 0)
        self.assertEqual(prod(range(1, 6)), 120)
        self.assertEqual(prod(range(1, 6), 1), 120)
        self.assertEqual(prod((), 1), 1)
        self.assertRaises(ValueError, prod, ())

    def test_frange(self) -> None:
        self.assertIterableAlmostEqual(frange(5), (0, 1, 2, 3, 4))
        self.assertIterableAlmostEqual(frange(1, 5), (1, 2, 3, 4))
        self.assertIterableAlmostEqual(frange(1, 5, 0.5), (1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5))

    def test_linspace(self) -> None:
        self.assertIterableAlmostEqual(linspace(0, 5), frange(0, 5, 0.05))
        self.assertIterableAlmostEqual(linspace(0, 5, 5), range(5))

    def test_interp(self) -> None:
        self.assertAlmostEqual(interp(1, 0, 2, 0, 3), 1.5)
        self.assertAlmostEqual(interp(1, 0, 1, 0, 3), 3)
        self.assertAlmostEqual(interp(2, 0, 1, 0, 6), 12)

    def test_series_sum(self) -> None:
        self.assertEqual(series_sum(-5), -15)
        self.assertEqual(series_sum(-5, 0), -15)
        self.assertEqual(series_sum(0, -5), -15)
        self.assertEqual(series_sum(5), 15)
        self.assertEqual(series_sum(0, 5), 15)
        self.assertEqual(series_sum(0, 5, 6), 15)
        self.assertEqual(series_sum(0, 5, 2), 5)
        self.assertEqual(series_sum(1, 1, 1), 1)


if __name__ == '__main__':
    main()



from typing import Optional


def af(i: float, n: float) -> float:
    """Calculates the sinking fund factor.

    :param i: The interest rate.
    :param n: The number of periods.
    :return: The calculated factor.
    """
    return i / ((1 + i) ** n - 1)


def ag(i: float, n: float) -> float:
    """Calculates the arithmetic gradient to annuity conversion factor.

    :param i: The interest rate.
    :param n: The number of periods.
    :return: The calculated factor.
    """
    return 1 / i - n / ((1 + i) ** n - 1)


def ap(i: float, n: float) -> float:
    """Calculates the capital recovery factor.

    :param i: The interest rate.
    :param n: The number of periods.
    :return: The calculated factor.
    """
    return i * (1 + i) ** n / ((1 + i) ** n - 1)


def fa(i: float, n: float) -> float:
    """Calculates the uniform series compound amount factor.

    :param i: The interest rate.
    :param n: The number of periods.
    :return: The calculated factor.
    """
    return ((1 + i) ** n - 1) / i


def fp(i: float, n: float) -> float:
    """Calculates the compound amount factor.

    :param i: The interest rate.
    :param n: The number of periods.
    :return: The calculated factor.
    """
    return (1 + i) ** n


def pa(i: float, n: float, g: Optional[float] = None) -> float:
    """If the geometric gradient rate is supplied, calculates the geometric gradient to present worth conversion factor.
       Otherwise, calculates the series present worth factor.

    :param i: The interest rate.
    :param n: The number of periods.
    :param g: The optional geometric gradient rate.
    :return: The calculated factor.
    """
    if g is None:
        return ((1 + i) ** n - 1) / (i * (1 + i) ** n)
    else:
        return pa((1 + i) / (1 + g) - 1, n) / (1 + g)


def pf(i: float, n: float) -> float:
    """Calculates the present worth factor.

    :param i: The interest rate.
    :param n: The number of periods.
    :return: The calculated factor.
    """
    return (1 + i) ** -n


def pp(i: float) -> float:
    """Calculates the perpetuity to present worth factor.

    :param i: The interest rate.
    :return: The calculated factor.
    """
    return 1 / i


def pg(i: float, n: float) -> float:
    """Calculates the arithmetic gradient to present worth factor.

    :param i: The interest rate.
    :param n: The number of periods.
    :return: The calculated factor.
    """
    return 1 / i ** 2 * (1 - (1 + i * n) / (1 + i) ** n)



from math import cos, exp, log, sin
from unittest import main

from auxiliary import ExtTestCase

from math2.calc import diff, euler, newton


class CalcTestCase(ExtTestCase):
    def test_diff(self) -> None:
        self.assertAlmostEqual(diff(lambda x: x ** 3 - 2 * x ** 2, 3), 3 * 3 ** 2 - 4 * 3)
        self.assertAlmostEqual(diff(lambda x: -3 * sin(2 * x), 5), -6 * cos(2 * 5))

    def test_newton(self) -> None:
        self.assertAlmostEqual(newton(lambda x: (x - 1) * (x + 5), 4), 1)
        self.assertAlmostEqual(newton(lambda x: exp(2 * x) - 10, 0.1), log(10) / 2)

    def test_euler(self) -> None:
        self.assertAlmostEqual(euler(lambda x: x, 0, 1), 0.5, 2)
        self.assertAlmostEqual(euler(lambda x: 2 * x ** 2, -1, 1), 4 / 3, 3)


if __name__ == '__main__':
    main()



from collections.abc import Iterable, Iterator
from functools import reduce
from itertools import chain
from operator import add, mul
from typing import Optional

from auxiliary import default

from math2.typing import _SLT, _T


def bind(value: _SLT, lower: _SLT, upper: _SLT) -> _SLT:
    """Binds the value by the given interval.

    :param value: The value.
    :param lower: The lower limit.
    :param upper: The upper limit.
    :return: The bound value.
    """
    if upper < lower:
        raise ValueError('Lower bound is greater than the upper bound')
    elif value < lower:
        return lower
    elif upper < value:
        return upper
    else:
        return value


def sum_(values: Iterable[_T], start: Optional[_T] = None) -> _T:
    """Calculates the sum of the elements in the iterable.

    :param values: The values.
    :param start: The optional start value.
    :return: The sum.
    """
    try:
        return reduce(add, values if start is None else chain((start,), values))
    except TypeError:
        raise ValueError('Invalid iterable')


def prod(values: Iterable[_T], start: Optional[_T] = None) -> _T:
    """Calculates the product of the elements in the iterable.

    :param values: The values.
    :param start: The optional start value.
    :return: The product.
    """
    try:
        return reduce(mul, values if start is None else chain((start,), values))
    except TypeError:
        raise ValueError('Invalid iterable')


def frange(start: float, stop: Optional[float] = None, step: Optional[float] = None) -> Iterator[float]:
    """Generates a range of floating point values.

    :param start: The start value.
    :param stop: The stop value.
    :param step: The step value.
    :return: The iterator of range values.
    """
    if stop is None:
        yield from frange(0, start, step)
    elif step is None:
        yield from frange(start, stop, 1)
    else:
        while start < stop:
            yield start
            start += step


def linspace(start: float, stop: float, n: int = 100) -> Iterator[float]:
    """Generates an iterator of values from start to stop with length of n.

    :param start: The start value.
    :param stop: The stop value.
    :param n: The number of values.
    :return: The linspace of arguments.
    """
    return frange(start, stop, (stop - start) / n)


def interp(x: float, x0: float, x1: float, y0: float, y1: float) -> float:
    """Interpolates the point between two given points.

    :param x: The point to interpolate.
    :param x0: The x coordinate of the first point.
    :param x1: The x coordinate of the second point.
    :param y0: The y coordinate of the first point.
    :param y1: The y coordinate of the second point.
    :return: The interpolated value.
    """
    return (x - x0) / (x1 - x0) * (y1 - y0) + y0


def series_sum(lo: int, hi: Optional[int] = None, n: Optional[int] = None) -> int:
    """Calculates the series sum of the interval.

    :param lo: The start value.
    :param hi: The optional end value.
    :param n: The number of elements, defaults to (end - start) + 1.
    :return: the series sum.
    """
    return (lo + default(hi, 0)) * default(n, abs(default(hi, 0) - lo) + 1) // 2



from collections.abc import Callable
from math import ceil, exp, inf, log
from unittest import TestCase, main

from auxiliary import ExtTestCase

from math2.calc import newton
from math2.econ import (Bond, CashFlow, CompoundInterest, ContinuousInterest, DblDeclBalDeprec, DeclBalDeprec,
                        EffectiveInterest, Mortgage, NominalInterest, Project, Relationship, SYDDeprec, SimpleInterest,
                        StrLineDeprec, SubperiodInterest, UPDeprec, fp, from_table, irr, pa, payback_period, pf, pg,
                        related_combinations, relationship)
from math2.misc import interp


class InterestTestCase(TestCase):
    def test_difference(self) -> None:
        r, p, t, s, c = 0.07, 24, 2020 - 1626, 685.92, 9066082143624.828

        self.assertAlmostEqual(p * SimpleInterest(r).to_factor(t), s)
        self.assertAlmostEqual(p * EffectiveInterest(r).to_factor(t), c)

    def test_comparison(self) -> None:
        self.assertLess(NominalInterest(0.06, 12).to_effective().rate, SubperiodInterest(0.063, 1).to_effective().rate)

    def test_consistency(self) -> None:
        nr, sc, t, f = 0.1, 4, 2.5, 1.2800845441963565
        counts = range(1, 366)

        interests = [
            EffectiveInterest((1 + nr / sc) ** sc - 1),
            ContinuousInterest(log((1 + nr / sc) ** sc)),
            NominalInterest(nr, sc),
            SubperiodInterest(nr / sc, sc),
        ]

        for interest in interests:
            self.assertAlmostEqual(interest.to_factor(t), f)
            self.assertAlmostEqual(interest.to_effective().to_factor(t), f)
            self.assertAlmostEqual(interest.to_continuous().to_factor(t), f)

            for count in counts:
                self.assertAlmostEqual(interest.to_nominal(count).to_factor(t), f)
                self.assertAlmostEqual(interest.to_subperiod(count).to_factor(t), f)

        self.assertAlmostEqual(NominalInterest(nr, sc).to_nominal().to_factor(t), f)
        self.assertAlmostEqual(NominalInterest(nr, sc).to_subperiod().to_factor(t), f)
        self.assertAlmostEqual(SubperiodInterest(nr / sc, sc).to_nominal().to_factor(t), f)
        self.assertAlmostEqual(SubperiodInterest(nr / sc, sc).to_subperiod().to_factor(t), f)


class InstrumentTestCase(ExtTestCase):
    def test_relationships(self) -> None:
        self.assertEqual(relationship([5000, 7000, 6000, 3000], 1000000), Relationship.INDEPENDENT)
        self.assertEqual(relationship([5000, 7000, 6000, 3000], 7000), Relationship.MUTUALLY_EXCLUSIVE)
        self.assertEqual(relationship([5000, 7000, 6000, 3000], 10000), Relationship.RELATED)

    def test_combinations(self) -> None:
        self.assertEqual(len(tuple(related_combinations([5000, 7000, 6000, 3000], 10000))), 8)

    def test_projects(self) -> None:
        self.assertAlmostEqual(Project(-20000, 4000 - 1000, 4000, 10).present_worth(EffectiveInterest(0.05)),
                               9680.783294664216)
        self.assertAlmostEqual(Project(-20000, 4000 - 1000, 4000, 10).annual_worth(EffectiveInterest(0.05)),
                               727.9268005526942)


class CashFlowTestCase(TestCase):
    def test_payback_period(self) -> None:
        self.assertAlmostEqual(payback_period([], 0), 0)
        self.assertAlmostEqual(payback_period([], 10), inf)
        self.assertAlmostEqual(payback_period([CashFlow(0, 100), CashFlow(1, 200), CashFlow(2, 300)], 450), 1.5)


class PS1TestCase(TestCase):
    def test_1(self) -> None:
        self.assertLess(8000 + 1200, 6800 + 2500)
        self.assertAlmostEqual(newton(lambda mv: 8000 + 1200 - (mv + 2500), 6800), 6700)

    def test_2(self) -> None:
        self.assertAlmostEqual(newton(lambda t: 7500 - t * 1700, 0), 4.411764705882353)
        self.assertAlmostEqual(newton(lambda t: 9000 - t * 2200, 0), 4.09090909)

    def test_3(self) -> None:
        self.assertAlmostEqual((1 + 0.05 / 1000000) ** 1000000, exp(0.05))

    def test_4(self) -> None:
        e = EffectiveInterest(0.034)
        p, t = 100000, 4

        self.assertAlmostEqual(e.to_nominal(2).rate, 0.03371581102178567)
        self.assertAlmostEqual(e.to_nominal(12).rate, 0.033481397886386155)
        self.assertAlmostEqual(e.to_nominal(365).rate, 0.03343630748129267)
        self.assertAlmostEqual(e.to_continuous().rate, 0.03343477608623742)

        self.assertAlmostEqual(p * e.to_nominal(2).to_factor(t), p * e.to_nominal(12).to_factor(t))
        self.assertAlmostEqual(p * e.to_nominal(12).to_factor(t), p * e.to_nominal(365).to_factor(t))
        self.assertAlmostEqual(p * e.to_nominal(365).to_factor(t), p * e.to_continuous().to_factor(t))
        self.assertAlmostEqual(p * e.to_continuous().to_factor(t), 114309.4552336)

    def test_5(self) -> None:
        fv, c, cd = 100, 0.023, 0.02

        self.assertAlmostEqual(ContinuousInterest(c).to_effective().rate, 0.02326653954721758)
        self.assertAlmostEqual(fv / ContinuousInterest(c).to_factor(0.5), 98.8565872247913)
        self.assertAlmostEqual(fv / ContinuousInterest(c).to_factor(0.75), 98.28979294344309)
        self.assertAlmostEqual(fv / ContinuousInterest(cd).to_factor(0.5), 99.0049833749168)
        self.assertAlmostEqual(fv / ContinuousInterest(cd).to_factor(0.75), 98.51119396030627)

    def test_6(self) -> None:
        self.assertAlmostEqual(EffectiveInterest(0.08).to_subperiod(12).rate, 0.00643403011000343)
        self.assertAlmostEqual(NominalInterest(0.035, 252).to_effective().rate, 0.03561719190449408)
        self.assertAlmostEqual(NominalInterest(0.04, 4).to_continuous().rate, 0.039801323412672354)
        self.assertAlmostEqual(CompoundInterest.from_factor(SubperiodInterest(0.015, 12).to_factor(1), 4)
                               .to_continuous().rate, 0.04466583748125169)
        self.assertAlmostEqual(CompoundInterest.from_factor(CompoundInterest.from_factor(
            NominalInterest(0.012, 3).to_factor(1), 1 / 4).to_factor(3), 1).to_nominal(6).rate, 0.1454477030768886)


class PS2TestCase(ExtTestCase):
    def test_1(self) -> None:
        factor = NominalInterest(0.15, 4).to_factor(4 / 12) * ContinuousInterest(0.11).to_factor(5 / 12)

        self.assertAlmostEqual(CompoundInterest.from_factor(factor, 9 / 12).rate, 0.134915472328168)

    def test_2(self) -> None:
        p, i, n = 100, 0.05, 10

        self.assertAlmostEqual(p * pa(i, n), p * ((1 + i) ** n - 1) / (i * (1 + i) ** n))

    def test_3(self) -> None:
        cases = [
            (1, 99.52, 0.05943811, 0.05773868),
            (4, 99.01, 0.03029791, 0.02984799),
            (8, 97.64, 0.03647384, 0.03582441),
            (12, 95.85, 0.04329682, 0.04238572),
        ]

        for m, p, a, b in cases:
            self.assertAlmostEqual(newton(lambda y: p * EffectiveInterest(y).to_factor(m / 12) - 100, 0), a)
            self.assertAlmostEqual(newton(lambda y: p * ContinuousInterest(y).to_factor(m / 12) - 100, 0), b)

        self.assertAlmostEqual(interp(7, 4, 8, 0.03029791, 0.03647384), 0.0349298575)
        self.assertAlmostEqual(interp(7, 4, 8, 0.02984799, 0.03582441), 0.034330305)

        self.assertAlmostEqual(newton(lambda y: 1.03029791 ** (4. / 12) * (1 + y) ** (8. / 12) - 1.04329682, 0),
                               0.04985765)

    def test_4(self) -> None:
        data = {
            2: 99.11,
            6: 98.72,
            9: 96.85,
            12: 95.97,
        }

        r: dict[float, float] = {
            t: newton(lambda y: p * ContinuousInterest(y).to_factor(t / 12) - 100, 0)
            for t, p in data.items()
        }

        r[4] = interp(4, 2, 6, r[2], r[6])
        r[5.5] = interp(5.5, 2, 6, r[2], r[6])
        r[7] = interp(7, 6, 9, r[6], r[9])
        r[10] = interp(10, 9, 12, r[9], r[12])

        self.assertAlmostEqual(100 * ContinuousInterest(r[2]).to_factor(-1. / 12), 99.55400544428134)
        self.assertAlmostEqual(100 * ContinuousInterest(r[4]).to_factor(-4. / 12), 98.68531340424114)
        self.assertAlmostEqual(100 * ContinuousInterest(r[5.5]).to_factor(-5.5 / 12), 98.66834503291024)
        self.assertAlmostEqual(100 * ContinuousInterest(r[7]).to_factor(-7 / 12), 98.18488742486127)
        self.assertAlmostEqual(100 * ContinuousInterest(r[10]).to_factor(-10 / 12), 96.54750684004732)
        self.assertAlmostEqual(2000 * pf(r[2], 1. / 12) + 500 * pf(r[4], 4. / 12) - 1200 * pf(r[7], 7. / 12) - 1000
                               * pf(r[10], 10. / 12) + 500 * pf(r[12], 1), 820.3872407904064)

    def test_5(self) -> None:
        i = 0.02
        options = (
            15500 * pa(i, 10),
            140000,
            155000 * pf(i, 5),
            170000 * pf(i, 10),
        )

        self.assertSequenceAlmostEqual(options, (139230.06759675476, 140000, 140388.27552363696, 139459.21097877636))
        self.assertEqual(max(range(4), key=options.__getitem__), 2)

    def test_6(self) -> None:
        i = 0.02
        options = (
            155000 * pf(i, 5),
            10000 * pa(i, 10) + 1500 * pg(i, 10),
        )

        self.assertAlmostEqual(options[1], 148258.50062422457)
        self.assertEqual(max(range(2), key=options.__getitem__), 1)

    def test_7(self) -> None:
        i = 0.02
        option = 10000 * pa(i, 10, g=0.05)

        self.assertAlmostEqual(option, 112086.97925088322)


class PS3TestCase(TestCase):
    def test_1(self) -> None:
        m = Mortgage.from_dtv(2995000, 0.2)
        i1 = NominalInterest(0.02, 2)
        i2 = NominalInterest(0.04, 2)
        p = m.payment(i1)

        self.assertAlmostEqual(p, 10145.891129693951)
        self.assertAlmostEqual(p * 12 * 3, 365252.08066898223)
        self.assertAlmostEqual(m.pay(i1, 5).payment(i2), 12128.043601452593)

    def test_2(self) -> None:
        i = NominalInterest(0.07, 2)

        self.assertAlmostEqual(Bond(100, 0, 2, 3 / 12).present_worth(i), 98.2946374365981)
        self.assertAlmostEqual(Bond(100, 0, 2, 5 / 12).present_worth(i), 97.17391685967232)
        self.assertAlmostEqual(Bond(100, 0, 2, 3).present_worth(i), 81.35006443077528)
        self.assertAlmostEqual(Bond.from_rate(100, 0.04, 2, 3).present_worth(i), 92.00717047033226)
        self.assertAlmostEqual(Bond.from_rate(100, 0.06, 2, 3.25).present_worth(i), 97.13753584095278)

    def test_3(self) -> None:
        self.assertAlmostEqual(
            newton(lambda y: Bond.from_rate(100, 0.07, 2, 3).present_worth(NominalInterest(y, 2)) - 100, 0.1), 0.07)
        self.assertAlmostEqual(
            Bond.from_rate(100, 0.04, 2, 3).present_worth(NominalInterest(0.05, 2)) + 100 * 0.04 / 2, 99.24593731921009)
        self.assertAlmostEqual(
            newton(lambda y: Bond.from_rate(100, 0.03, 2, 2.25).present_worth(NominalInterest(y, 2)) - 100, 0.1), 0.03)
        self.assertAlmostEqual(
            Bond.from_rate(100, 0.07, 2, 2.25).present_worth(NominalInterest(0.05, 2)), 104.20662940110009)
        self.assertAlmostEqual(newton(
            lambda c: Bond.from_rate(100, c, 2, 2.25).present_worth(NominalInterest(0.03, 2)) - 114, 0.1), 0.09481118)

    def test_4(self) -> None:
        pass

    def test_5(self) -> None:
        y = newton(
            lambda y_: Bond.from_rate(100, 0.07, 2, 7.5).present_worth(NominalInterest(y_, 2)) * fp(y_ / 2, 0.5) - 108,
            0.1)

        b: Callable[[float], float] = lambda c_: Bond.from_rate(1000, c_, 2, 9).present_worth(NominalInterest(y, 2))
        c = ceil(newton(lambda c_: 9500000 / 2 - (4400 * b(c_)), 0.1) / 0.0025) * 0.0025
        self.assertAlmostEqual(c, 0.0725)
        self.assertAlmostEqual(4400 * b(c), 4802235.185695931)
        c = ceil(newton(lambda c_: 9500000 / 2 / (1 - 0.008) - (4400 * b(c_)), 0.1) / 0.0025) * 0.0025
        self.assertAlmostEqual(c, 0.0725)
        self.assertAlmostEqual(4400 * b(c) * (1 - 0.008), 4763817.304210364)

    def test_6(self) -> None:
        i = NominalInterest(0.060755, 2)

        self.assertAlmostEqual(Mortgage.from_down(500000, 50000).payment(i), 2899.3558026129626)
        self.assertAlmostEqual(Mortgage.from_down(500000, 50000).pay(i, 3, 700).payment(i), 3490.3113416458878)
        self.assertLess(Mortgage.from_down(500000, 50000, 25).pay(i, 3).principal,
                        Mortgage.from_down(500000, 50000, 25).pay(i, 3, 700).principal)


class PS6TestCase(ExtTestCase):
    def test_1(self) -> None:
        data = ((-41000, 6100, 7),
                (-32000, 6700, 7),
                (-28000, 5700, 5),
                (-28000, 12600, 5),
                (-36000, 9000, 7),
                (-27000, 10600, 6),
                (-53000, 6700, 5),
                (-50000, 15000, 6),
                (-32000, 6900, 7),
                (-42000, 14600, 5))

        irr_set = tuple(map(lambda d: irr(Project(d[0], d[1], 0, d[2]).cash_flows(), 0).rate, data))

        self.assertSequenceAlmostEqual(irr_set, (
            0.010261108929599895,
            0.10584583010815002,
            0.005929015028005828,
            0.3494328573992243,
            0.16326709023510008,
            0.31754169406374866,
            -0.13571830650187303,
            0.1990541470961173,
            0.114956469240095,
            0.2178733729868983,
        ))

        points = sorted(range(len(data)), key=irr_set.__getitem__, reverse=True)

        cost = 0
        marr = 0.0

        for point in points:
            cost -= data[point][0]

            if cost > 100000:
                break

            marr = irr_set[point]

        self.assertAlmostEqual(marr, 0.2178733729868983)

    def test_2(self) -> None:
        self.assertEqual(from_table(
            ((),
             (0.17,),
             (0.14, 0.075),
             (0.19, 0.209, 0.286),
             (0.2, 0.127, 0.257, 0.229),
             (0.18, 0.177, 0.192, 0.158, 0.117),
             (0.13, 0.128, 0.132, 0.106, 0.081, 0.062)),
            0.12,
        ), 4)

        self.assertEqual(from_table(
            ((),
             (0.14,),
             (0.20, 0.29),
             (0.24, 0.32, 0.36),
             (0.21, 0.24, 0.22, 0.11),
             (0.17, 0.18, 0.15, 0.08, 0.06),
             (0.17, 0.18, 0.16, 0.12, 0.13, 0.19)),
            0.12,
        ), 3)

    def test_3(self) -> None:
        table = ((),
                 (0.1096,),
                 (0.132, 0.286),
                 (0.1205, 0.17, -0.058),
                 (0.1293, 0.189, 0.112, 0.228),
                 (0.1286, 0.177, 0.112, 0.187, 0.113),
                 (0.1113, 0.113, 0.079, 0.094, 0.069, 0.063))

        self.assertEqual(from_table(table, 0.04), 6)
        self.assertEqual(from_table(table, 0.06), 6)
        self.assertEqual(from_table(table, 0.08), 5)
        self.assertEqual(from_table(table, 0.10), 5)
        self.assertEqual(from_table(table, 0.12), 2)
        self.assertEqual(from_table(table, 0.14), 0)

    def test_4(self) -> None:
        pass

    def test_5(self) -> None:
        pass


class PS7TestCase(ExtTestCase):
    def test_1(self) -> None:
        basis, salvage, life = 92000, 19000, 4

        self.assertIterableAlmostEqual(StrLineDeprec(basis, salvage, life).books, (92000, 73750, 55500, 37250, 19000))
        self.assertIterableAlmostEqual(DeclBalDeprec(basis, salvage, life).books,
                                       (92000, 62019.64424847585, 41809.08992073374, 28184.618296048306, 19000))
        self.assertIterableAlmostEqual(DeclBalDeprec.from_rate(basis, life, 0.25).books,
                                       (92000, 69000, 51750, 38812.5, 29109.375))
        self.assertIterableAlmostEqual(DblDeclBalDeprec(basis, salvage, life).books,
                                       (92000, 46000, 23000, 11500, 5750))
        self.assertIterableAlmostEqual(SYDDeprec(basis, salvage, life).books, (92000, 62800, 40900, 26300, 19000))
        self.assertIterableAlmostEqual(UPDeprec(basis, salvage, (37000, 37000, 32000, 30000)).books,
                                       (92000, 72139.70588235295, 52279.41176470589, 35102.94117647059, 19000))

    def test_2(self) -> None:
        basis, salvage, life, t = 210000, 10000, 20, 6

        self.assertAlmostEqual(StrLineDeprec(basis, salvage, life).book(t), 150000)
        self.assertAlmostEqual(DeclBalDeprec(basis, salvage, life).book(t), 84246.81795174461)
        self.assertAlmostEqual(DeclBalDeprec.from_rate(basis, life, 0.2).book(t), 55050.24)
        self.assertAlmostEqual(DblDeclBalDeprec(basis, salvage, life).book(t), 111602.61)

        self.assertAlmostEqual(StrLineDeprec(basis, salvage, life).amount(t), 10000)
        self.assertAlmostEqual(DeclBalDeprec(basis, salvage, life).amount(t), 13852.157371177402)
        self.assertAlmostEqual(DeclBalDeprec.from_rate(basis, life, 0.2).amount(t), 13762.56)
        self.assertAlmostEqual(DblDeclBalDeprec(basis, salvage, life).amount(t), 12400.29)

    def test_3(self) -> None:
        basis, life, t, book = 145000, 8, 6, 57000

        self.assertAlmostEqual(newton(lambda x: StrLineDeprec(basis, x, life).book(6) - book, 0), 27666.66666666667)
        self.assertAlmostEqual(newton(lambda x: DeclBalDeprec(basis, x, life).book(6) - book, 0), 41755.1908917986)

    def test_4(self) -> None:
        basis, salvage, life = 110000, 25000, 4

        self.assertIterableAlmostEqual(StrLineDeprec(basis, salvage, life).books, (110000, 88750, 67500, 46250, 25000))
        self.assertIterableAlmostEqual(DeclBalDeprec(basis, salvage, life).books,
                                       (110000, 75950.3039160202, 52440.44240850757, 36207.88671287913, 25000))
        self.assertIterableAlmostEqual(DeclBalDeprec.from_rate(basis, life, 0.35).books,
                                       (110000, 71500, 46475, 30208.75, 19635.6875))
        self.assertIterableAlmostEqual(DblDeclBalDeprec(basis, salvage, life).books,
                                       (110000, 55000, 27500, 13750, 6875))
        self.assertIterableAlmostEqual(SYDDeprec(basis, salvage, life).books, (110000, 76000, 50500, 33500, 25000))
        self.assertIterableAlmostEqual(UPDeprec(basis, salvage, (80000, 65000, 50000, 35000)).books,
                                       (110000, 80434.78260869565, 56413.043478260865, 37934.78260869565, 25000))

    def test_5(self) -> None:
        basis, salvage, life, t1, t2 = 23000, 4000, 7, 4, 5

        self.assertAlmostEqual(StrLineDeprec(basis, salvage, life).book(t1), 12142.857142857143)
        self.assertAlmostEqual(DeclBalDeprec(basis, salvage, life).book(t1), 8465.096723059049)
        self.assertAlmostEqual(SYDDeprec(basis, salvage, life).book(t1), 8071.4285714285725)
        self.assertAlmostEqual(UPDeprec(basis, salvage, (50, 60, 40, 20, 10, 15, 5)).book(t1), 6850)

        self.assertAlmostEqual(StrLineDeprec(basis, salvage, life).amount(t2), 2714.285714285714)
        self.assertAlmostEqual(DeclBalDeprec(basis, salvage, life).amount(t2), 1871.7191438152927)
        self.assertAlmostEqual(SYDDeprec(basis, salvage, life).amount(t2), 2035.7142857142856)
        self.assertAlmostEqual(UPDeprec(basis, salvage, (50, 60, 40, 20, 10, 15, 5)).amount(t2), 950)

    def test_6(self) -> None:
        basis, salvage, life, t = 2500000, 200000, 10, 4

        self.assertAlmostEqual(StrLineDeprec(basis, salvage, life).book(t), 1580000)
        self.assertAlmostEqual(DeclBalDeprec(basis, salvage, life).book(t), 910282.1015130404)
        self.assertAlmostEqual(DblDeclBalDeprec(basis, salvage, life).book(t), 1024000)
        self.assertAlmostEqual(SYDDeprec(basis, salvage, life).book(t), 1078181.8181818182)

        self.assertAlmostEqual(StrLineDeprec(basis, salvage, life).amount(t), 230000)
        self.assertAlmostEqual(DeclBalDeprec(basis, salvage, life).amount(t), 261554.3542830096)
        self.assertAlmostEqual(DblDeclBalDeprec(basis, salvage, life).amount(t), 256000)
        self.assertAlmostEqual(SYDDeprec(basis, salvage, life).amount(t), 292727.2727272727)

    def test_7(self) -> None:
        deprec = DeclBalDeprec.from_rate(5000, 5, 0.15)

        self.assertAlmostEqual(deprec.cap_gain(3000), 0)
        self.assertAlmostEqual(deprec.recap_deprec(3000), 781.4734375)
        self.assertAlmostEqual(deprec.loss_on_disp(3000), 0)

        self.assertAlmostEqual(deprec.cap_gain(2000), 0)
        self.assertAlmostEqual(deprec.recap_deprec(2000), 0)
        self.assertAlmostEqual(deprec.loss_on_disp(2000), 218.5265625)

        self.assertAlmostEqual(deprec.cap_gain(6000), 1000)
        self.assertAlmostEqual(deprec.recap_deprec(6000), 2781.4734375)
        self.assertAlmostEqual(deprec.loss_on_disp(6000), 0)


if __name__ == '__main__':
    main()



from unittest import main

from auxiliary import ExtTestCase

from math2.stats import range_, shuffled, trimmed_mean


class StatsTestCase(ExtTestCase):
    value_sets = ((0.32, 0.53, 0.28, 0.37, 0.47, 0.43, 0.36, 0.42, 0.38, 0.43),
                  (0.26, 0.43, 0.47, 0.49, 0.52, 0.75, 0.79, 0.86, 0.62, 0.46),
                  (7.07, 7.00, 7.10, 6.97, 7.00, 7.03, 7.01, 7.01, 6.98, 7.08),
                  (1., 2., 3., 4., 5.))

    def test_trimmed_mean(self) -> None:
        self.assertIterableAlmostEqual(map(lambda seq: trimmed_mean(seq, 0.1), self.value_sets),
                                       (0.39750, 0.56625, 7.0225, 3))

    def test_range(self) -> None:
        self.assertAlmostEqual(range_(map(float, range(5))), 4)
        self.assertIterableAlmostEqual(map(range_, self.value_sets), (0.25, 0.6, 0.13, 4))

    def test_shuffled(self) -> None:
        a = list(range(1000))

        self.assertNotEqual(shuffled(a), a)


if __name__ == '__main__':
    main()



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from enum import Enum
from itertools import chain
from typing import Optional

from auxiliary import iindex, retain_iter

from math2.econ.cashflows import CashFlow
from math2.econ.factors import af, ap, fa, fp, pa
from math2.econ.interests import CompoundInterest
from math2.misc import frange


class Instrument(ABC):
    """Instrument is the abstract base class for all financial instruments."""

    @abstractmethod
    def cash_flows(self, interest: CompoundInterest) -> Iterator[CashFlow]:
        """Calculates the cash flows of this instrument at the given interest value.

        :param interest: The interest value.
        :return: The cash flows.
        """
        pass

    @abstractmethod
    def present_worth(self, interest: CompoundInterest) -> float:
        """Calculates the present worth of this instrument at the given interest value.

        :param interest: The interest value.
        :return: The present worth.
        """
        pass

    @abstractmethod
    def annual_worth(self, interest: CompoundInterest) -> float:
        """Calculates the annual worth of this instrument at the given interest value.

        :param interest: The interest value.
        :return: The annual worth.
        """
        pass


class Bond(Instrument):
    """Bond is the class for bonds."""

    def __init__(self, face: float, coupon: float, period_count: float, maturity: float):
        self.face = face
        self.coupon = coupon
        self.period_count = period_count
        self.maturity = maturity

    @property
    def period(self) -> float:
        """
        :return: The period of this bond.
        """
        return 1 / self.period_count

    def cash_flows(self, interest: Optional[CompoundInterest] = None) -> Iterator[CashFlow]:
        return chain((CashFlow(t, self.coupon) for t in frange(self.period, self.maturity + self.period, self.period)),
                     (CashFlow(self.maturity, self.face),))

    def present_worth(self, interest: CompoundInterest) -> float:
        return self.coupon * pa(interest.to_subperiod(self.period_count).rate, self.maturity * self.period_count) \
               + self.face / interest.to_factor(self.maturity)

    def annual_worth(self, interest: CompoundInterest) -> float:
        return self.present_worth(interest) * ap(interest.to_effective().rate, self.maturity)

    @classmethod
    def from_rate(cls, face: float, rate: float, period_count: float, maturity: float) -> Bond:
        """Creates the bond from the coupon rate.

        :param face: The face value.
        :param rate: The coupon rate.
        :param period_count: The period count.
        :param maturity: The maturity of the bond.
        :return: The created bond.
        """
        return Bond(face, face * rate / period_count, period_count, maturity)


class Mortgage(Instrument):
    """Mortgage is the class for mortgages."""

    def __init__(self, principal: float, frequency: float = 12, amortization: float = 25):
        self.principal = principal
        self.frequency = frequency
        self.amortization = amortization

    def cash_flows(self, interest: CompoundInterest) -> Iterator[CashFlow]:
        payment = self.payment(interest)

        return (CashFlow(t, payment) for t in frange(0, self.amortization, 1 / self.frequency))

    def present_worth(self, interest: Optional[CompoundInterest] = None) -> float:
        return 0

    def annual_worth(self, interest: Optional[CompoundInterest] = None) -> float:
        return 0

    def payment(self, interest: CompoundInterest) -> float:
        """Calculates the payment value with respect to the given interest value.

        :param interest: The interest value.
        :return: The payment.
        """
        return self.principal * ap(interest.to_subperiod(self.frequency).rate, self.frequency * self.amortization)

    def pay(self, interest: CompoundInterest, term: float, payment: Optional[float] = None) -> Mortgage:
        """Creates a new mortgage instance assuming payments were made.

        :param interest: The interest value.
        :param term: The term during which payments were made.
        :param payment: The optional payment made, defaults to payment with respect to the interest.
        :return: The next mortgage instance.
        """
        if payment is None:
            return self.pay(interest, term, self.payment(interest))
        else:
            return Mortgage(
                self.principal * interest.to_factor(term) - payment
                * fa(interest.to_subperiod(self.frequency).rate, term * self.frequency),
                self.frequency, self.amortization - term,
            )

    @classmethod
    def from_down(cls, value: float, down: float, frequency: float = 12, amortization: float = 25) -> Mortgage:
        """Constructs the mortgage instance from the down payment value.

        :param value: The total value.
        :param down: The down payment.
        :param frequency: The frequency of payments, defaults to 12.
        :param amortization: The amortization of the mortgage, defaults to 25.
        :return: The constructed mortgage instance.
        """
        return cls(value - down, frequency, amortization)

    @classmethod
    def from_dtv(cls, value: float, dtv: float, frequency: float = 12, amortization: float = 25) -> Mortgage:
        """Constructs the mortgage instance from the down payment to value percentage.

        :param value: The total value.
        :param dtv: The down payment to value percentage.
        :param frequency: The frequency of payments, defaults to 12.
        :param amortization: The amortization of the mortgage, defaults to 25.
        :return: The constructed mortgage instance.
        """
        return cls.from_down(value, value * dtv, frequency, amortization)

    @classmethod
    def from_ltv(cls, value: float, ltv: float, frequency: float = 12, amortization: float = 25) -> Mortgage:
        """Constructs the mortgage instance from the loan payment to value percentage.

        :param value: The total value.
        :param ltv: The loan payment to value percentage.
        :param frequency: The frequency of payments, defaults to 12.
        :param amortization: The amortization of the mortgage, defaults to 25.
        :return: The constructed mortgage instance.
        """
        return cls(value * ltv, frequency, amortization)


class Project(Instrument):
    """Project is the class for projects."""

    def __init__(self, initial: float, annuity: float, final: float, life: float):
        self.initial = initial
        self.annuity = annuity
        self.final = final
        self.life = life

    def cash_flows(self, interest: Optional[CompoundInterest] = None) -> Iterator[CashFlow]:
        return chain((CashFlow(0, self.initial),), (CashFlow(t + 1, self.annuity) for t in frange(self.life)),
                     (CashFlow(self.life, self.final),))

    def present_worth(self, interest: CompoundInterest) -> float:
        rate = interest.to_effective().rate

        return self.initial + self.annuity * pa(rate, self.life) + self.final * fp(rate, self.life)

    def annual_worth(self, interest: CompoundInterest) -> float:
        rate = interest.to_effective().rate

        return self.initial * ap(rate, self.life) + self.annuity + self.final * af(rate, self.life)

    def rep_present_worth(self, interest: CompoundInterest, total_life: float) -> float:
        """Calculates the repeated present worth of this project given total life.

        :param interest: The interest value.
        :param total_life: The total life during which the project is repeated.
        :return: The repeated present worth of this project.
        """
        present_worth = self.present_worth(interest)

        return sum(present_worth * interest.to_factor(t) for t in frange(0, total_life, self.life))


class Relationship(Enum):
    """Relationship is the enum class for all relationships of alternative projects."""
    INDEPENDENT = 0
    MUTUALLY_EXCLUSIVE = 1
    RELATED = 2


@retain_iter
def relationship(values: Iterable[float], budget: float) -> Relationship:
    """Determines the relationship of values with respect to the budget.

    :param values: The values.
    :param budget: The budget.
    :return: The relationship.
    """
    if sum(values) <= budget:
        return Relationship.INDEPENDENT
    elif sum(sorted(values)[:2]) <= budget:
        return Relationship.RELATED
    else:
        return Relationship.MUTUALLY_EXCLUSIVE


def related_combinations(values: Iterable[float], budget: float) -> Iterator[Iterator[int]]:
    """Gets the combinations of the related values given their values and the budget.

    :param values: The values of the projects.
    :param budget: The budget.
    :return: The combinations of possible projects that can be chosen.
    """
    if values := tuple(values):
        i = len(values) - 1
        chosen = related_combinations(values[:i], budget - values[i]) if values[i] <= budget else ()
        skipped = related_combinations(values[:i], budget)

        return chain((chain(sub_combination, [i]) for sub_combination in chosen), skipped)
    else:
        return iter((iter(()),))


def from_table(table: Iterable[Iterable[float]], marr: float) -> int:  # TODO ACCEPT irrs
    """Selects the project with respect to the given table of internal rate of returns and marr.

    :param table: The table of internal rate of returns.
    :param marr: The minimum acceptable rate of return.
    :return: The best project.
    """
    x = 0

    for i, row in enumerate(table):
        if i and iindex(row, x) > marr:
            x = i

    return x



from setuptools import find_packages, setup

with open('README.rst', 'r') as long_description_file:
    long_description = long_description_file.read()

setup(
    name='math2',
    version='0.0.2.dev2',
    author='Juho Kim',
    author_email='juho-kim@outlook.com',
    description='A Python package for various math functions',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/AussieSeaweed/math2',
    packages=find_packages(),
    package_data={'math2': ('py.typed',)},
    classifiers=(
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ),
    python_requires='>=3.9',
    install_requires='auxiliary',
)



from unittest import main

from auxiliary import ExtTestCase

from math2.linalg import Vector, full, ones, replaced, zeros


class VectorTestCase(ExtTestCase):
    def test_init(self) -> None:
        self.assertIterableAlmostEqual(Vector(range(10)), range(10))

        self.assertIterableAlmostEqual(zeros(10), (0,) * 10)
        self.assertIterableAlmostEqual(ones(10), (1,) * 10)
        self.assertIterableAlmostEqual(full(10, 1 / 3), (1 / 3,) * 10)
        self.assertIterableAlmostEqual(replaced(ones(10), 5, 0), (1,) * 5 + (0,) + (1,) * 4)

    def test_ops(self) -> None:
        a, b = Vector(range(10)), Vector(range(10))

        self.assertIterableAlmostEqual(+a, a)
        self.assertIterableAlmostEqual(-a, (-x for x in range(10)))

        self.assertIterableAlmostEqual(a + b, (x + x for x in range(10)))
        self.assertIterableAlmostEqual(a - b, zeros(10))
        self.assertIterableAlmostEqual(a * 3, (x * 3 for x in range(10)))
        self.assertIterableAlmostEqual(3 * a, (3 * x for x in range(10)))
        self.assertIterableAlmostEqual(a / 3, (x / 3 for x in range(10)))
        self.assertAlmostEqual(a @ b, sum(x * x for x in range(10)))

        a += b
        self.assertIterableAlmostEqual(a, (x + x for x in range(10)))
        a -= b
        self.assertIterableAlmostEqual(a, b)
        a *= 3
        self.assertIterableAlmostEqual(a, (x * 3 for x in range(10)))
        a /= 3
        self.assertIterableAlmostEqual(a, range(10))

        self.assertRaises(ValueError, a.__add__, range(5))
        self.assertRaises(ValueError, a.__sub__, range(5))


if __name__ == '__main__':
    main()



from itertools import chain

from math2.linalg.data import Vector


def zeros(size: int) -> Vector:
    return Vector(0 for _ in range(size))


def ones(size: int) -> Vector:
    return Vector(1 for _ in range(size))


def full(size: int, value: float) -> Vector:
    return Vector(value for _ in range(size))


def deleted(vector: Vector, index: int) -> Vector:
    return Vector(x for i, x in enumerate(vector) if i != index)


def inserted(vector: Vector, index: int, value: float) -> Vector:
    return Vector(chain(vector[:index], (value,), vector[index:]))


def replaced(vector: Vector, index: int, value: float) -> Vector:
    return Vector(value if index == i else x for i, x in enumerate(vector))



from collections.abc import Callable

from math2.misc import frange


def diff(f: Callable[[float], float], x: float, eps: float = 1e-7) -> float:
    """Perform numerical differentiation on the supplied function.

    :param f: The function to be differentiated.
    :param x: The point of differentiation.
    :param eps: The accuracy.
    :return: The derivative at the point.
    """
    return (f(x + eps) - f(x - eps)) / (2 * eps)


def newton(f: Callable[[float], float], x: float, eps: float = 1e-7) -> float:
    """Solves the root of the supplied function with the Newton's method.
.
    :param f: The function to solve.
    :param x: The initial guess.
    :param eps: The accuracy.
    :return: The root coordinate.
    """
    while eps < abs(y := f(x)):
        x -= y / diff(f, x, eps)
    else:
        return x


def euler(f: Callable[[float], float], xlo: float, xhi: float, n: float = 100) -> float:
    """Integrates the function between two given bounds with the Euler's method.

    :param f: The function to integrate.
    :param xlo: The lower bound.
    :param xhi: The upper bound.
    :param n: The number of evaluation points.
    :return: The numerical integral of the function.
    """
    dx = (xhi - xlo) / n

    return sum(dx * f(x) for x in frange(xlo, xhi, dx))



from __future__ import annotations

from collections.abc import Iterable, MutableSequence
from typing import Union, overload

from auxiliary import const_len


class Vector(MutableSequence[float]):
    def __init__(self, values: Iterable[float] = ()):
        self.__values = list(values)

    @overload
    def __getitem__(self, i: int) -> float:
        ...

    @overload
    def __getitem__(self, s: slice) -> Vector:
        ...

    def __getitem__(self, i: Union[int, slice]) -> Union[float, Vector]:
        return self.__values[i] if isinstance(i, int) else Vector(self.__values[i])

    @overload
    def __setitem__(self, i: int, o: float) -> None:
        ...

    @overload
    def __setitem__(self, s: slice, o: Iterable[float]) -> None:
        ...

    def __setitem__(self, i: Union[int, slice], o: Union[float, Iterable[float]]) -> None:
        if isinstance(i, int) and isinstance(o, float):
            self.__values[i] = o
        elif isinstance(i, slice) and isinstance(o, Iterable):
            self.__values[i] = o
        else:
            raise TypeError('Invalid argument types')

    @overload
    def __delitem__(self, i: int) -> None:
        ...

    @overload
    def __delitem__(self, i: slice) -> None:
        ...

    def __delitem__(self, i: Union[int, slice]) -> None:
        del self.__values[i]

    def __len__(self) -> int:
        return len(self.__values)

    def __pos__(self) -> Vector:
        return Vector(self)

    def __neg__(self) -> Vector:
        return Vector(-x for x in self)

    @const_len
    def __add__(self, other: Iterable[float]) -> Vector:
        return Vector(x + y for x, y in zip(self, other))

    @const_len
    def __sub__(self, other: Iterable[float]) -> Vector:
        return Vector(x - y for x, y in zip(self, other))

    def __mul__(self, other: float) -> Vector:
        return Vector(x * other for x in self)

    def __rmul__(self, other: float) -> Vector:
        return self * other

    def __truediv__(self, other: float) -> Vector:
        return Vector(x / other for x in self)

    @const_len
    def __matmul__(self, other: Iterable[float]) -> float:
        return sum(x * y for x, y in zip(self, other))

    @const_len
    def __iadd__(self, other: Iterable[float]) -> Vector:
        self.__values = (self + other).__values

        return self

    @const_len
    def __isub__(self, other: Iterable[float]) -> Vector:
        self.__values = (self - other).__values

        return self

    def __imul__(self, other: float) -> Vector:
        self.__values = (self * other).__values

        return self

    def __itruediv__(self, other: float) -> Vector:
        self.__values = (self / other).__values

        return self

    def insert(self, index: int, value: float) -> None:
        self.__values.insert(index, value)



from math2.econ.cashflows import CashFlow, disc, disc_payback_period, irr, npv, payback_period
from math2.econ.depreciations import DblDeclBalDeprec, DeclBalDeprec, Deprec, SYDDeprec, StrLineDeprec, UPDeprec
from math2.econ.factors import af, ag, ap, fa, fp, pa, pf, pg, pp
from math2.econ.instruments import (Bond, Instrument, Mortgage, Project, Relationship, from_table, related_combinations,
                                    relationship)
from math2.econ.interests import (CompoundInterest, ContinuousInterest, EffectiveInterest, Interest,
                                  MultipleCompoundInterest, NominalInterest, SimpleInterest, SubperiodInterest)

__all__ = ('CashFlow', 'disc', 'disc_payback_period', 'irr', 'npv', 'payback_period', 'DblDeclBalDeprec',
           'DeclBalDeprec', 'Deprec', 'SYDDeprec', 'StrLineDeprec', 'UPDeprec', 'af', 'ag', 'ap', 'fa', 'fp', 'pa',
           'pf', 'pg', 'pp', 'Bond', 'Instrument', 'Mortgage', 'Project', 'Relationship', 'from_table',
           'related_combinations', 'relationship', 'CompoundInterest', 'ContinuousInterest', 'EffectiveInterest',
           'Interest', 'MultipleCompoundInterest', 'NominalInterest', 'SimpleInterest', 'SubperiodInterest')



from collections.abc import Iterable
from functools import total_ordering
from itertools import accumulate
from math import inf
from typing import Any

from auxiliary import retain_iter, windowed

from math2.calc import newton
from math2.econ.interests import EffectiveInterest, Interest
from math2.misc import interp


@total_ordering
class CashFlow:
    """CashFlow is the class for cash flows."""

    def __init__(self, time: float, amount: float):
        self.time = time
        self.amount = amount

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, CashFlow):
            return self.time == other.time
        else:
            return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, CashFlow):
            return self.time < other.time
        else:
            return NotImplemented


def disc(cash_flow: CashFlow, interest: Interest) -> CashFlow:
    """Discounts the cash flow from a copied value.

    :param cash_flow: The cash flow.
    :param interest: The interest at which is discounted.
    :return: The discounted copy of the cash flow.
    """
    return CashFlow(cash_flow.time, cash_flow.amount / interest.to_factor(cash_flow.time))


def npv(cash_flows: Iterable[CashFlow], interest: Interest) -> float:
    """Calculates the net present value of the supplied cash flows at the interest.

    :param cash_flows: The cash flows.
    :param interest: The interest.
    :return: The net present value.
    """
    return sum(disc(cash_flow, interest).amount for cash_flow in cash_flows)


@retain_iter
def irr(cash_flows: Iterable[CashFlow], initial_guess: float) -> EffectiveInterest:
    """Calculates the internal rate of return using the initial guess.

    :param cash_flows: The cash flows.
    :param initial_guess: The initial guess.
    :return: The internal rate of return.
    """
    return EffectiveInterest(newton(lambda i: npv(cash_flows, EffectiveInterest(i)), initial_guess))


def payback_period(cash_flows: Iterable[CashFlow], cost: float) -> float:
    """Calculates the payback period of the cash flows.

    :param cash_flows: The cash flows.
    :param cost: The cost to pay back.
    :return: The payback period.
    """
    prefix_sums = tuple(accumulate(cash_flow.amount for cash_flow in sorted(cash_flows)))

    if cost <= 0 or prefix_sums and cost <= prefix_sums[0]:
        return 0
    else:
        for i, (x, y) in enumerate(windowed(prefix_sums, 2)):
            if y >= cost:
                return interp(cost, x, y, i, i + 1)
        else:
            return inf


def disc_payback_period(cash_flows: Iterable[CashFlow], cost: float, interest: Interest) -> float:
    """Calculates the discounted payback period of the cash flows at the given interest value.

    :param cash_flows: The cash flows.
    :param cost: The cost to pay back.
    :param interest: The interest at which the cash flows are discounted.
    :return: The payback period.
    """
    return payback_period(map(lambda cash_flow: disc(cash_flow, interest), cash_flows), cost)



from __future__ import annotations

from abc import ABC, abstractmethod
from math import exp, inf, log
from typing import Optional

from math2.econ.factors import fp


class Interest(ABC):
    """Interest is the abstract base class for all interests."""

    def __init__(self, rate: float):
        self.rate = rate

    @abstractmethod
    def to_factor(self, time: float) -> float:
        """Converts this interest to the factor at the given time period.

        :param time: The time period.
        :return: The converted factor.
        """
        pass

    @classmethod
    @abstractmethod
    def from_factor(cls, factor: float, time: float) -> Interest:
        """Converts the factor at a time period to an interest value.

        :param factor: The factor.
        :param time: The time period.
        :return: The converted interest value.
        """
        pass


class SimpleInterest(Interest):
    """SimpleInterest is the class for simple interests."""

    def to_factor(self, time: float) -> float:
        return 1 + self.rate * time

    @classmethod
    def from_factor(cls, factor: float, time: float) -> SimpleInterest:
        return SimpleInterest((factor - 1) / time)


class CompoundInterest(Interest, ABC):
    """CompoundInterest is the abstract base class for all compound interests."""

    @abstractmethod
    def to_effective(self) -> EffectiveInterest:
        """Converts this interest value to an effective interest value.

        :return: The converted interest value.
        """
        pass

    @abstractmethod
    def to_continuous(self) -> ContinuousInterest:
        """Converts this interest value to a continuous interest value.

        :return: The converted interest value.
        """
        pass

    @abstractmethod
    def to_nominal(self, subperiod_count: float) -> NominalInterest:
        """Converts this interest value to a nominal interest value.

        :param subperiod_count: The number of subperiods of the converted interest value.
        :return: The converted interest value.
        """
        pass

    @abstractmethod
    def to_subperiod(self, subperiod_count: float) -> SubperiodInterest:
        """Converts this interest value to a subperiod interest value.

        :param subperiod_count: The number of subperiods of the converted interest value.
        :return: The converted interest value.
        """
        pass

    @classmethod
    def from_factor(cls, factor: float, time: float) -> CompoundInterest:
        return EffectiveInterest(factor ** (1 / time) - 1)


class EffectiveInterest(CompoundInterest):
    """EffectiveInterest is the class for effective interests."""

    def to_factor(self, time: float) -> float:
        return fp(self.rate, time)

    def to_effective(self) -> EffectiveInterest:
        return self

    def to_continuous(self) -> ContinuousInterest:
        return ContinuousInterest(log(self.rate + 1))

    def to_nominal(self, subperiod_count: float) -> NominalInterest:
        return NominalInterest(subperiod_count * ((self.rate + 1) ** (1 / subperiod_count) - 1), subperiod_count)

    def to_subperiod(self, subperiod_count: float) -> SubperiodInterest:
        return SubperiodInterest((self.rate + 1) ** (1 / subperiod_count) - 1, subperiod_count)


class MultipleCompoundInterest(CompoundInterest, ABC):
    """MultipleCompoundInterest is the abstract base class for all multiple compounding interests."""

    def __init__(self, rate: float, subperiod_count: float):
        super().__init__(rate)

        self.subperiod_count = subperiod_count

    @property
    def subperiod(self) -> float:
        """
        :return: The subperiod of this multiple compound interest value.
        """
        return 1 / self.subperiod_count


class NominalInterest(MultipleCompoundInterest):
    """NominalInterest is the class for nominal interests."""

    def to_factor(self, time: float) -> float:
        return fp(self.rate / self.subperiod_count, self.subperiod_count * time)

    def to_effective(self) -> EffectiveInterest:
        return EffectiveInterest(fp(self.rate / self.subperiod_count, self.subperiod_count) - 1)

    def to_continuous(self) -> ContinuousInterest:
        return self.to_effective().to_continuous()

    def to_nominal(self, subperiod_count: Optional[float] = None) -> NominalInterest:
        return self if subperiod_count is None else self.to_effective().to_nominal(subperiod_count)

    def to_subperiod(self, subperiod_count: Optional[float] = None) -> SubperiodInterest:
        if subperiod_count is None:
            return SubperiodInterest(self.rate / self.subperiod_count, self.subperiod_count)
        else:
            return self.to_effective().to_subperiod(subperiod_count)


class SubperiodInterest(MultipleCompoundInterest):
    """SubperiodInterest is the class for subperiod interests."""

    def to_factor(self, time: float) -> float:
        return fp(self.rate, self.subperiod_count * time)

    def to_effective(self) -> EffectiveInterest:
        return EffectiveInterest(fp(self.rate, self.subperiod_count) - 1)

    def to_continuous(self) -> ContinuousInterest:
        return self.to_effective().to_continuous()

    def to_nominal(self, subperiod_count: Optional[float] = None) -> NominalInterest:
        if subperiod_count is None:
            return NominalInterest(self.rate * self.subperiod_count, self.subperiod_count)
        else:
            return self.to_effective().to_nominal(subperiod_count)

    def to_subperiod(self, subperiod_count: Optional[float] = None) -> SubperiodInterest:
        return self if subperiod_count is None else self.to_effective().to_subperiod(subperiod_count)


class ContinuousInterest(MultipleCompoundInterest):
    """ContinuousInterest is the class for continuous interests."""

    def __init__(self, rate: float):
        super().__init__(rate, inf)

    def to_factor(self, time: float) -> float:
        return exp(self.rate) ** time

    def to_effective(self) -> EffectiveInterest:
        return EffectiveInterest(exp(self.rate) - 1)

    def to_continuous(self) -> ContinuousInterest:
        return self

    def to_nominal(self, subperiod_count: float) -> NominalInterest:
        return self.to_effective().to_nominal(subperiod_count)

    def to_subperiod(self, subperiod_count: float) -> SubperiodInterest:
        return self.to_effective().to_subperiod(subperiod_count)
